# -*- coding: utf-8 -*-
class Blueprint:
    """
    一个简单的蓝图,只实现了延迟添加路由的功能
    """

    def __init__(self, app, prefixurl=None):
        self.app = app
        self.prefixurl = prefixurl
        self.addtasklist = []

    def add(self, urlregex, handler, handlername=None, methods=["GET"]):
        if self.prefixurl:
            if not urlregex.startswith(r"^/"):
                urlregex = r"^/" + self.prefixurl + "/" + urlregex[1:]
            else:
                urlregex = r"^" + self.prefixurl + urlregex[1:]
        args = (urlregex, handler, handlername, methods)
        self.addtasklist.append(args)

    def register(self):
        for args in self.addtasklist:
            self.app.route.add(*args)



# -*- coding: utf-8 -*-
import re
from datetime import datetime

# RFC6265 http://stackoverflow.com/questions/1969232/allowed-characters-in-cookies
LegalCharsValue = r"[A-Za-z0-9!#\$%&'\*\+-\.^_`\|~]*"
LegalCharsHeader = r"[A-Za-z0-9!#$%&'\(\)\*\+-\./:<=>?@\[\]^_`\{\|\}~]*"

Translator = {n: '\\{:03o}'.format(n) for n in range(256) if not re.match(LegalCharsHeader, chr(n))}
Translator.update({
    ord('"'): '\\"',
    ord('\\'): '\\\\',
})


def _quote(header):
    # print(header)
    if not header or re.match(r"^[A-Za-z0-9]*$", header):
        return header
    else:
        return '"' + header.translate(Translator) + '"'


class Cookie(dict):
    """
    A class to hold ONE (key, value) pair.
    一个简化了的http.cookies.morsel
    """
    _reserved = {
        "expires": "expires",
        "path": "Path",
        "comment": "Comment",
        "domain": "Domain",
        "max-age": "Max-Age",
        "secure": "Secure",
        "httponly": "HttpOnly",
        "version": "Version",
    }

    _flags = {'secure',
              'httponly'
              }

    def __init__(self, key, value):
        # Set defaults
        if key in self._reserved:
            raise KeyError("Invalid Key")
        if not re.match(r"^[A-Za-z0-9]*$", key):
            raise KeyError("Illlegal key")
        # Set default attributes
        super().__init__()
        for _key in self._reserved:
            super().__setitem__(_key, "")
        for _key in self._flags:
            super().__setitem__(_key, False)
        self.key = key
        self.value = value

    def __setitem__(self, key, value):
        if key not in self._reserved:
            raise KeyError("Unknown cookie property")
        return super().__setitem__(key, value)

    def out(self):
        out = ["{}={}".format(self.key, _quote(self.value))]
        for key, value in sorted(self.items()):
            if key == "max-age" and isinstance(value, int):
                out.append("{}={}".format(self._reserved[key], value))
            elif key == "expires" and isinstance(value, datetime):
                out.append("{}={}".format(
                    self._reserved[key],
                    value.strftime("%a, %d-%b-%Y %T GMT")
                ))
            elif key in self._flags and self[key]:
                out.append(self._reserved[key])
            else:
                if value:
                    out.append("{}={}".format(self._reserved[key], value))
        return "; ".join(out)


class CookieJar(dict):
    def __init__(self):
        super(CookieJar, self).__init__()

    def __setitem__(self, key, value):
        cookie = self.get(key)
        if not cookie:
            cookie = Cookie(key, value)
        return super().__setitem__(key, cookie)


def cookiesparser(cookies):
    cookies = cookies.strip(" ")
    cookiesdict = {}
    cookielist = cookies.split(";")[:-1] if cookies[-1] == ";" else  cookies.split(";")
    for cookie in cookielist:
        pos = cookie.index("=")
        value = cookie[pos + 1:].strip(" ") if ";" not in cookie[pos + 1:] else cookie[pos + 1:cookies.index(";")]
        cookiesdict[cookie[:pos].strip(" ")] = value
    return cookiesdict

    # cookies="sdfsdfsdf=sfsdfsfsdf; _xsrf=2|972dbe8e|478161f590ba6dca7d86f24c1b6ed0d9|1492335497; "
    # print(cookiesparser(cookies))



# -*- coding: utf-8 -*-
import datetime
import json

from Ksana.cookies import CookieJar

STATUSDICT = {
    100: "Continue",
    200: "OK",
    302: "FOUND",
    401: "UNAUTHORIZED",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",

}


class Response:
    charset = "utf-8"
    default_status = 200
    default_mimetype = "text/plain"
    default_version = "1.1"

    def __init__(self, body=None, headers=None, version=None, status=None, content_type=None, keep_live=None,
                 encode=None):
        self.version = version if version else self.default_version
        self.status = status if status else self.default_status
        self.date = datetime.datetime.utcnow().strftime('%a, %d %b %Y %H:%M:%S GMT')
        self.content_type = content_type if content_type else self.default_mimetype
        self.keep_live = keep_live if keep_live else  ""
        self.headers = headers if headers else {}
        self.cookies = CookieJar()
        self.body = body if body else ""
        self.encode = encode

    def make_response(self):
        if self.encode:
            return self.encode_make_response()
        headerofkeeplive = "Keep-Alive: timeout={}\r\n".format(self.keep_live) if self.keep_live else ""
        baseresponse = "HTTP/{version:} {status:} {statusresponse:}\r\n" \
                       "Date: {date:}\r\n" \
                       "Content-Type: {content_type:}\r\n" \
                       "Content-Length: {length:}\r\n" \
                       "Connction:{keeplive:}\r\n{keepliveset:}" \
                       "{headers:}" \
                       "{cookies:}" \
                       "\r\n{body}".format(
            version=self.version,
            status=self.status,
            statusresponse=STATUSDICT[self.status],
            date=self.date,
            content_type=self.content_type + ((" ;" + self.charset) if self.content_type == "text/plain" else ""),
            length=len(self.body),
            keeplive='keep-alive' if self.keep_live else 'close',
            keepliveset=headerofkeeplive,
            headers="\r\n".join(
                "{}: {}".format(key, value) for key, value in self.headers.items()) + "\r\n" if self.headers else "",
            cookies="\r\n".join("Set-Cookie: {}".format(self.cookies[cookie].out()) for cookie in self.cookies) + (
                " \r\n" if self.cookies else ""),
            body=self.body,
        )
        return baseresponse.encode("utf-8")

    def set_cookie(self, key, value):
        return self.cookies.__setitem__(key, value)

    def delete_cookie(self, key, request, path='/', domain=None):
        print("start delete cookies")
        if not request.cookiedict.get(key):
            return
        value = request.cookiedict[key]
        self.set_cookie(key, value)
        self.cookies[key]["path"] = path
        if domain:
            self.cookies[key][domain] = domain
        self.cookies[key]["expires"] = "0"
        self.cookies[key]["max-age"] = "0"
        print(self.cookies)
        print("end")

    def encode_make_response(self):
        headerofkeeplive = "Keep-Alive: timeout={}\r\n".format(self.keep_live) if self.keep_live else ""
        baseresponse = "HTTP/{version:} {status:} {statusresponse:}\r\n" \
                       "Date: {date:}\r\n" \
                       "Content-Type: {content_type:}\r\n" \
                       "Content-Length: {length:}\r\n" \
                       "Content-Encoding: {encode}\r\n" \
                       "Connction:{keeplive:}\r\n{keepliveset:}" \
                       "{headers:}" \
                       "{cookies:}".format(
            version=self.version,
            status=self.status,
            statusresponse=STATUSDICT[self.status],
            date=self.date,
            content_type=self.content_type + ((" ;" + self.charset) if self.content_type == "text/plain" else ""),
            length=len(self.body),
            keeplive='keep-alive' if self.keep_live else 'close',
            keepliveset=headerofkeeplive,
            headers="\r\n".join(
                "{}: {}".format(key, value) for key, value in self.headers.items()) + "\r\n" if self.headers else "",
            cookies="\r\n".join("Set-Cookie: {}".format(self.cookies[cookie].out()) for cookie in self.cookies) + (
                " \r\n" if self.cookies else ""),
            encode=self.encode
        )
        baseresponse = baseresponse.encode("utf-8")
        return baseresponse + b"\r\n" + self.body


class Responsejson(Response):
    def __init__(self, body=None, headers=None, version=None, status=None, keep_live=None, encode=None):
        body = json.dumps(body)
        content_type = "application/json"
        self.encode = encode
        super().__init__(body=body, headers=headers, version=version, status=status, content_type=content_type,
                         keep_live=keep_live)


class Responsefile(Response):
    def __init__(self, filename=None, headers=None, version=None, status=None, content_type=None, keep_live=None,
                 encode=None):
        if not encode:
            with open(filename, "r") as file:
                body = file.read()
            content_type = content_type or "text/plain; charset=utf-8"
        else:
            with open(filename, "rb") as file:
                body = file.read()
            content_type = content_type

        self.encode = encode
        super().__init__(body=body, headers=headers, version=version, status=status, content_type=content_type,
                         keep_live=keep_live, encode=encode)


class Redirctresponse(Response):
    def __init__(self, to, body="", headers=None, version=None, status=302, content_type=None, keep_live=None,
                 encode=None):
        print("start redirct")
        headers = {"Location": to}
        super().__init__(body=body, headers=headers, version=version, status=status, content_type=content_type,
                         keep_live=keep_live, encode=encode)


class Responseheader(Response):
    def __init__(self, to, body="", headers=None, version=None, status=200, content_type=None, keep_live=None,
                 encode=None):
        super().__init__(body=body, headers=headers, version=version, status=status, content_type=content_type,
                         keep_live=keep_live, encode=encode)

    def make_response(self):
        headerofkeeplive = "Keep-Alive: timeout={}\r\n".format(self.keep_live) if self.keep_live else ""
        baseresponse = "HTTP/{version:} {status:} {statusresponse:}\r\n" \
                       "Date: {date:}\r\n" \
                       "Content-Type: {content_type:}\r\n" \
                       "Content-Length: {length:}\r\n" \
                       "Connction:{keeplive:}\r\n{keepliveset:}" \
                       "{headers:}" \
                       "{cookies:}" \
            .format(
            version=self.version,
            status=self.status,
            statusresponse=STATUSDICT[self.status],
            date=self.date,
            content_type=self.content_type + ((" ;" + self.charset) if self.content_type == "text/plain" else ""),
            length=len(self.body),
            keeplive='keep-alive' if self.keep_live else 'close',
            keepliveset=headerofkeeplive,
            headers="\r\n".join(
                "{}: {}".format(key, value) for key, value in self.headers.items()) + "\r\n" if self.headers else "",
            cookies="\r\n".join("Set-Cookie: {}".format(self.cookies[cookie].out()) for cookie in self.cookies) + (
                " \r\n" if self.cookies else ""),
        )
        return baseresponse.encode("utf-8")



# -*- coding: utf-8 -*-
import asyncio
import traceback
from inspect import getmodulename, currentframe, getframeinfo
from inspect import isawaitable
from signal import signal,SIGTERM, SIGINT

from Ksana.errors import ErrorHandler, KsanaError
from .baseserver import start_server
from .route import Route


class Ksana:
    def __init__(self, name=None, route=None, error_handle=None):
        self.name = name if name else getmodulename(getframeinfo(currentframe()).filename)
        self.route = route if route else Route()
        self.error_handle = error_handle if error_handle else ErrorHandler()
        #每个app维持一个队列,暂不考虑多进程的情况
        self.loop=asyncio.new_event_loop()
        self.beforeresponse = []
        self.afterresponse = []

    async def handle_request(self,request,write):
        try:
            handler, args, kwargs = self.route.get(request)
            for before in self.beforeresponse:
                before(request)
            response = handler(request, *args, **kwargs)
            for after in self.afterresponse:
                after(request, response)
            if isawaitable(response):
                print("isawaitable")
                response = await response
                print("adada")
        except KsanaError as e:
                #暂时只实现了404,405,500,没有定义异常视图的情况下,默认500
                response = self.error_handle.response(request, e)
        except Exception as e:
            print(repr(e))
            traceback.print_exc()

        write(response)

    def run(self,address,
                 request_time=10,
                 backlog=100,
                 reuse_port=False):
        signal(SIGTERM, self.loop.stop)
        signal(SIGINT, self.loop.stop)
        start_server(address=address,
                     requesthandle=self.handle_request,
                     request_time=request_time,
                     backlog=backlog,
                     reuse_port=reuse_port)



# -*- coding: utf-8 -*-
import asyncio
from pprint import pprint
from signal import SIGINT, SIGTERM

from httptools import HttpRequestParser

from Ksana.request import Request


class BaseServer(asyncio.Protocol):
    def __init__(self, loop, requesthandle, toggle, connections=set(), request_timeout=10):
        self.loop = loop
        self.requesthandle = requesthandle
        self.toggle = toggle
        self.connections = connections
        self.request_timeout = request_timeout

        self.timehandle = None
        self.requesthandletask = None
        # 默认请求参数
        self.ip = None
        self.parse = None
        self.url = None
        self.headers = {}
        self.body = None
        self.httpversion = None
        self.method = None
        self.request = None
        self.contentlength = 0
        self.keep_alive = False
        # print("enter init")

    #######################
    # 连接建立
    def connection_made(self, transport):
        # 将本连接加入全局的连接集合
        print("connect start")
        self.connections.add(self)
        self.transport = transport
        self.ip = transport.get_extra_info("peername")
        # self.timehandle = self.loop.call_later(self.request_timeout, self.teardown)

    def connection_lost(self, exc):
        print("connect end")
        self.connections.remove(self)
        # self.timehandle.cancel()
        self.clean()

    ########################
    # 解析数据
    def data_received(self, data):
        pprint(data.decode("utf-8"))
        self.contentlength += len(data)
        if not self.parse:
            # HttpRequestParser响应当前对象的以下方法
            # - on_url(url:byte)
            # - on_header(name: bytes, value: bytes)
            # - on_headers_complete()
            # - on_body(body: bytes)
            # - on_message_complete()
            # get_http_version(self) -> str
            # def should_keep_alive(self) -> bool:
            self.parse = HttpRequestParser(self)
        try:
            self.parse.feed_data(data)
        except HttpRequestParser as e:
            pass


    def on_url(self, url):
        self.url = url

    def on_header(self, name, value):
        self.headers[name] = value

    def on_headers_complete(self):
        self.method = self.parse.get_method()
        self.httpversion = self.parse.get_http_version()
        self.keep_alive = self.parse.should_keep_alive()
        self.request = Request(self.ip, self.url, self.headers, self.method, self.httpversion,
                               self.request_timeout if self.keep_alive else None)

    def on_body(self, body):
        # print(body)
        self.body = body
        self.request.setbody(body)

    def on_message_complete(self):
        print("parser complete")
        self.requesthandletask = self.loop.create_task(self.requesthandle(self.request, self.write))

    #########################
    # 返回响应
    def write(self, response):
        print("start write")
        self.transport.write(response.make_response())
        print("end write")
        keep_alive = self.keep_alive and not self.toggle[0]
        if not keep_alive:
            self.transport.close()
        else:
            self.clean()

    ########################
    # 清理或者关闭连接
    def clean(self):
        self.requsettask = None
        self.parse = None
        self.url = None
        self.headers = {}
        self.body = None
        self.httpversion = None
        self.methods = None
        self.request = None
        self.contentlength = 0
        self.keep_alive = False

    def teardown(self):
        if not self.parse:
            self.transport.close()
            return True
        return False


def start_server(address,
                 requesthandle,
                 request_time=10,
                 loop=None,
                 sock=None,
                 backlog=100,
                 reuse_port=False
                 ):
    if not loop:
        loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    toggle = [False]
    connections = set()
    server = lambda: BaseServer(loop, requesthandle, toggle, connections, request_time)
    server_future = loop.create_server(server, address[0], address[1], reuse_port=reuse_port, sock=sock,
                                       backlog=backlog)
    server_running = loop.run_until_complete(server_future)
    loop.add_signal_handler(SIGINT, loop.stop)
    loop.add_signal_handler(SIGTERM, loop.stop)
    try:
        print("server start at {}:{}".format(*address))
        loop.run_forever()
    finally:
        server_running.close()
        loop.run_until_complete(server_running.wait_closed())
        toggle[0] = True
        for connection in connections:
            connection.teardown()
        loop.close()



# -*- coding: utf-8 -*-
import re
from collections import namedtuple

from Ksana.errors import MethodNotAllowed, NotFound

route = namedtuple("route", ["endpoint", "methods"])




class Route:
    """
    采用类似djngo的正则路由匹配
    """
    def __init__(self):
        self.routes = {}
        self.endtoview = {}

    def add(self, urlregex, handler, handlername=None, methods=["GET"]):
        '''
        :param urlregex: 匹配路径
        :param handler: 拦截函数
        :param handlername: 拦截函数名
        :param methods: 支持的方法
        '''
        if not urlregex.startswith(r"^/"):
            urlregex = r"^/" + urlregex[1:]
        if not handlername:
            handlername = handler.__name__
        self.routes[urlregex] = route(handlername, methods)
        self.endtoview[handlername] = handler

    def get(self, request):
        return self._get(request.url, request.method)

    def _get(self, url, method):
        if not url.startswith("/"):
            url += "/"
        for regex in self.routes:
            matchgroup = re.match(regex, url)
            if matchgroup:
                route = self.routes[regex]
                handlername = route.endpoint
                methods = route.methods
                if method in methods:
                    kwargs = matchgroup.groupdict()
                    # urlregex中匿名捕获和非匿名捕获的顺序,必须按照先位置参数,后关键词参数的顺序排列
                    args = matchgroup.groups()[:-len(kwargs)] if kwargs else matchgroup.groups()
                    handler = self.endtoview[handlername]
                    return handler, args, kwargs
                raise MethodNotAllowed
        raise NotFound

# def printadc():
#     print("abc")
# def printanything(thing):
#     print(thing)
# def printspecialthing(thing):
#     print(thing)
# myroute=Route()
# myroute.add(r"^/hello$",printadc)
# handler,args,kwargs=myroute.get("/hello","GET")
# handler(*args,*kwargs)
# myroute.add(r"^/hello/(.+?)$",printanything)
# handler,args,kwargs=myroute.get("/hello/adad","GET")
# handler(*args,*kwargs)
# myroute.add(r"^/helloworld/(?P<thing>.+?)$",printspecialthing)
# handler,args,kwargs=myroute.get("/helloworld/somethingspecial","GET")
# handler(*args,*kwargs)



#-*- coding: utf-8 -*-
from Ksana.response import Response
class KsanaError(Exception):
    pass


class NotFound(KsanaError):
    status = 404

class MethodNotAllowed(KsanaError):
    status = 405

class ServerError(KsanaError):
    status = 500

class ErrorHandler:
    def __init__(self):
        self.errors=dict()

    def add(self,exception,handler):
        self.errors[exception]=handler

    def response(self,request,exception):
        errorhandeler=self.errors.get(exception,self.defaulthandler)

        response=errorhandeler(request,exception)
        return response

    def defaulthandler(self,request,exception):
        if issubclass(type(exception),KsanaError):
            return Response(status=exception.status)



# -*- coding: utf-8 -*-
import uuid

global SessionDict
SessionDict = {"1": "2"}


class Usersession(dict):
    def __init__(self, uid):
        super(Usersession, self).__init__()
        self.modified = False
        self.uid = uid


class SessionInterface:
    @classmethod
    def open(self, request):
        print("open start")
        cookiesdict = request.cookiedict
        if cookiesdict:
            uid = cookiesdict.get("uid")
        print("open", SessionDict)
        if not cookiesdict or not uid or not SessionDict.get(uid):
            uid = uuid.uuid4().hex
            usersession = SessionDict[uid] = Usersession(uid)
            usersession.modified = True
        else:
            usersession = SessionDict[uid]
        request.usersession = usersession
        print(SessionDict)
        print("end start")

    @classmethod
    def save(self, request, response):
        if not request.usersession.modified:
            return
        uid = request.usersession.uid
        response.set_cookie("uid", uid)
        print("save", SessionDict)



#-*- coding: utf-8 -*-
from setuptools import setup, find_packages
setup(
    name='ksana',
    version='0.0.2.2',
    description='A sample Python web framework',
    url='https://github.com/zhangguyu6/ksana',
    author='Guyu Zhang',
    author_email='zhangguyu6@gmail.com',
    license='MIT',
    packages=find_packages(exclude=['test*']),
    install_requires=['httptools>=0.0.9',],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='sample web asyncio framework',
)


#-*- coding: utf-8 -*-
import re
from urllib.parse import parse_qs

from httptools import parse_url

from Ksana.cookies import cookiesparser


class Request:
    def __init__(self, ip, url, headers, method, version, keep_alive=None):
        self.ip = ip
        self.query_string = ""
        self.args = {}
        self.rawurl=parse_url(url)
        self.url=self.rawurl.path.decode("utf-8")
        if self.rawurl.query:
            self.query_string = self.rawurl.query.decode("utf-8")
        if self.query_string:
            self.args = parsertodict(self.query_string)

        self.headers={key.decode("utf-8"):value.decode("utf-8") for key,value in  headers.items()}
        self.method=method.decode("utf-8")
        self.version=version
        self.keep_alive=keep_alive
        self.session = None
        self.usersession = None
        self.cookiedict = {}
        self.absurl = "http://" + self.headers['Host'] + self.url + self.query_string
        self._formdata = {}
        if self.headers.get("Cookie"):
            self.cookiedict = cookiesparser(self.headers.get("Cookie"))
        self.body = ""


    def setbody(self,body):
        self.body=body.decode("utf-8")

    @property
    def formdata(self):
        if not self.headers.get("Content-Type"):
            return {}
        if self._formdata or not is_form(self.headers["Content-Type"]):
            return self._formdata
        if "multipart/form-data" in self.headers["Content-Type"]:
            self._formdata = parsermulformdata(self.headers["Content-Type"], self.body)
            return self._formdata
        if "application/x-www-form-urlencoded" in self.headers["Content-Type"]:
            self._formdata = parserurlencodeformdata(self.body)
            return self._formdata
        return {}


def parsertodict(query_string):
    qsdict = parse_qs(query_string)
    if qsdict:
        return {key: value if len(value) > 1 else value[0] for key, value in qsdict.items()}
    else:
        return {key[:-1]: "" for key in query_string.split("&")}


def is_form(contenttype):
    formlist = ["multipart/form-data", "application/x-www-form-urlencoded"]
    for i in formlist:
        if i in contenttype:
            return True
    return False


def parsermulformdata(contenttype, body):
    """
    Content-Type: multipart/form-data; boundary=----WebKitFormBoundary7MA4YWxkTrZu0gW
    ------WebKitFormBoundary7MA4YWxkTrZu0gW
    Content-Disposition: form-data; name="key1"

    v1
    ------WebKitFormBoundary7MA4YWxkTrZu0gW
    Content-Disposition: form-data; name="key2"

    v2
    -----WebKitFormBoundary7MA4YWxkTrZu0gW--

    """
    boundary = contenttype[contenttype.index("=") + 1:]
    low = body.index(boundary)
    high = body.rindex(boundary)
    formdata = body[low:high].split(boundary)
    formdatadict = {}
    for data in formdata:
        data = data.strip()
        result = re.match(r".*name=\"(?P<name>.*)\"\r\n\r\n(?P<data>.*)", data)
        if result:
            formdatadict[result.groupdict()["name"]] = result.groupdict()["data"]
    return formdatadict


def parserurlencodeformdata(body):
    """
    key-value pair
    """
    if "------WebKitFormBoundary" in body:
        contenttype = "=" + body[:body.index("\r\n")]
        return parsermulformdata(contenttype, body)
    return parsertodict(body)
